﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Core.Views;
using HeuristicLab.Data;
using HeuristicLab.MainForm;
using HeuristicLab.Optimization;
using HeuristicLab.Problems.DataAnalysis;
using HeuristicLab.Visualization;
using Rectangle = HeuristicLab.Visualization.Rectangle;

namespace HeuristicLab.VariableInteractionNetworks.Views {
  [View("Variable Interaction Network")]
  [Content(typeof(RunCollection), IsDefaultView = false)]

  public sealed partial class RunCollectionVariableInteractionNetworkView : ItemView {
    public RunCollectionVariableInteractionNetworkView() {
      InitializeComponent();
      ConfigureNodeShapes();
    }

    public new RunCollection Content {
      get { return (RunCollection)base.Content; }
      set {
        if (value != null && value != Content) {
          base.Content = value;
        }
      }
    }

    private static VariableInteractionNetwork BuildNetworkFromSolutionQualities(RunCollection runs, double threshold, bool useBestRunsPerTarget = false) {
      var nodes = new Dictionary<string, IVertex>();
      var vn = new VariableInteractionNetwork();
      var targets = runs.GroupBy(x => ((IRegressionProblemData)x.Parameters["ProblemData"]).TargetVariable).ToList();
      var targetQualities = new Dictionary<string, double>();
      var targetInputs = new Dictionary<string, List<string>>();

      if (useBestRunsPerTarget) {
        foreach (var target in targets) {
          var bestRun = target.OrderBy(x => ((DoubleValue)x.Results["Best training solution quality"]).Value).First();
          var bestQuality = ((DoubleValue)bestRun.Results["Best training solution quality"]).Value;
          var pd = (IRegressionProblemData)bestRun.Parameters["ProblemData"];
          if (threshold > bestQuality) continue; // skip if quality is below the treshold
          targetQualities[target.Key] = bestQuality;
          targetInputs[target.Key] = pd.AllowedInputVariables.ToList();
        }
      } else {
        foreach (var target in targets) {
          var avgQuality = CalculateAverageQuality(new RunCollection(target));
          if (threshold > avgQuality) continue;
          targetQualities[target.Key] = avgQuality;
          var pd = (IRegressionProblemData)target.First().Parameters["ProblemData"];
          targetInputs[target.Key] = pd.AllowedInputVariables.ToList();
        }
      }

      foreach (var ti in targetQualities) {
        var target = ti.Key;
        var variables = targetInputs[ti.Key];
        var quality = ti.Value;
        IVertex targetNode;

        if (!nodes.TryGetValue(target, out targetNode)) {
          targetNode = new VariableNetworkNode { Label = target };
          vn.AddVertex(targetNode);
          nodes[target] = targetNode;
        }

        IVertex variableNode;
        if (variables.Count > 0) {
          var variableList = new List<string>(variables);
          variableList.Add(target);
          var junctionLabel = Concatenate(variableList);
          IVertex junctionNode;
          if (!nodes.TryGetValue(junctionLabel, out junctionNode)) {
            junctionNode = new JunctionNetworkNode { Label = string.Empty };
            vn.AddVertex(junctionNode);
            nodes[junctionLabel] = junctionNode;
          }
          IArc arc;
          foreach (var v in variables) {
            var impact = quality;
            if (!nodes.TryGetValue(v, out variableNode)) {
              variableNode = new VariableNetworkNode { Label = v };
              vn.AddVertex(variableNode);
              nodes[v] = variableNode;
            }
            arc = new Arc(variableNode, junctionNode) { Weight = impact };
            vn.AddArc(arc);
          }
          arc = new Arc(junctionNode, targetNode) { Weight = junctionNode.InArcs.Sum(x => x.Weight) };
          vn.AddArc(arc);
        } else {
          foreach (var v in variables) {
            var impact = quality;
            if (!nodes.TryGetValue(v, out variableNode)) {
              variableNode = new VariableNetworkNode { Label = v };
              vn.AddVertex(variableNode);
              nodes[v] = variableNode;
            }
            var arc = new Arc(variableNode, targetNode) { Weight = impact };
            vn.AddArc(arc);
          }
        }
      }

      return vn;
    }

    private static VariableInteractionNetwork BuildNetworkFromVariableImpacts(RunCollection runs, string qualityResultName, bool maximization, string impactsResultName, double threshold, bool useBestRunsPerTarget = false) {
      var nodes = new Dictionary<string, IVertex>();
      var vn = new VariableInteractionNetwork();
      var targets = runs.GroupBy(x => ((IRegressionProblemData)x.Parameters["ProblemData"]).TargetVariable).ToList();

      var targetImpacts = new Dictionary<string, Dictionary<string, double>>();

      if (useBestRunsPerTarget) {
        var bestRunsPerTarget = maximization ?
          targets.Select(x => x.OrderBy(y => ((DoubleValue)y.Results[qualityResultName]).Value).Last()) :
          targets.Select(x => x.OrderBy(y => ((DoubleValue)y.Results[qualityResultName]).Value).First());

        foreach (var run in bestRunsPerTarget) {
          var pd = (IRegressionProblemData)run.Parameters["ProblemData"];
          var target = pd.TargetVariable;
          var impacts = (DoubleMatrix)run.Results[impactsResultName];
          targetImpacts[target] = impacts.RowNames.Select((x, i) => new { Name = x, Index = i }).ToDictionary(x => x.Name, x => impacts[x.Index, 0]);
        }
      } else {
        foreach (var target in targets) {
          var averageImpacts = CalculateAverageImpacts(new RunCollection(target), impactsResultName);
          targetImpacts[target.Key] = averageImpacts;
        }
      }

      foreach (var ti in targetImpacts) {
        var target = ti.Key;
        var variableImpacts = ti.Value;
        IVertex targetNode;

        var variables = variableImpacts.Keys.Where(x => variableImpacts[x] >= threshold).ToList();
        if (variables.Count == 0) continue;

        if (!nodes.TryGetValue(target, out targetNode)) {
          targetNode = new VariableNetworkNode { Label = target };
          vn.AddVertex(targetNode);
          nodes[target] = targetNode;
        }

        IVertex variableNode;
        if (variables.Count > 1) {
          var variableList = new List<string>(variables) { target };
          var junctionLabel = Concatenate(variableList);
          IVertex junctionNode;
          if (!nodes.TryGetValue(junctionLabel, out junctionNode)) {
            junctionNode = new JunctionNetworkNode { Label = string.Empty };
            vn.AddVertex(junctionNode);
            nodes[junctionLabel] = junctionNode;
          }
          IArc arc;
          foreach (var v in variables) {
            var impact = variableImpacts[v];
            if (!nodes.TryGetValue(v, out variableNode)) {
              variableNode = new VariableNetworkNode { Label = v };
              vn.AddVertex(variableNode);
              nodes[v] = variableNode;
            }
            arc = new Arc(variableNode, junctionNode) { Weight = impact };
            vn.AddArc(arc);
          }
          arc = new Arc(junctionNode, targetNode) { Weight = junctionNode.InArcs.Sum(x => x.Weight) };
          vn.AddArc(arc);
        } else {
          foreach (var v in variables) {
            var impact = variableImpacts[v];
            if (!nodes.TryGetValue(v, out variableNode)) {
              variableNode = new VariableNetworkNode { Label = v };
              vn.AddVertex(variableNode);
              nodes[v] = variableNode;
            }
            var arc = new Arc(variableNode, targetNode) { Weight = impact };
            vn.AddArc(arc);
          }
        }
      }
      return vn;
    }

    private static double CalculateAverageQuality(RunCollection runs) {
      var pd = (IRegressionProblemData)runs.First().Parameters["ProblemData"];
      var target = pd.TargetVariable;
      var inputs = pd.AllowedInputVariables;

      if (!runs.All(x => {
        var problemData = (IRegressionProblemData)x.Parameters["ProblemData"];
        return target == problemData.TargetVariable && inputs.SequenceEqual(problemData.AllowedInputVariables);
      })) {
        throw new ArgumentException("All runs must have the same target and inputs.");
      }
      return runs.Average(x => ((DoubleValue)x.Results["Best training solution quality"]).Value);
    }

    private static Dictionary<string, double> CalculateAverageImpacts(RunCollection runs, string resultName) {
      var pd = (IRegressionProblemData)runs.First().Parameters["ProblemData"];
      var target = pd.TargetVariable;
      var inputs = pd.AllowedInputVariables.ToList();

      var impacts = inputs.ToDictionary(x => x, x => 0d);

      // check if all the runs have the same target and same inputs
      if (!runs.All(x => {
        var problemData = (IRegressionProblemData)x.Parameters["ProblemData"];
        return target == problemData.TargetVariable && inputs.SequenceEqual(problemData.AllowedInputVariables);
      })) {
        throw new ArgumentException("All runs must have the same target and inputs.");
      }

      foreach (var run in runs) {
        var impactsMatrix = (DoubleMatrix)run.Results[resultName];

        int i = 0;
        foreach (var v in impactsMatrix.RowNames) {
          impacts[v] += impactsMatrix[i, 0];
          ++i;
        }
      }

      foreach (var v in inputs) {
        impacts[v] /= runs.Count;
      }

      return impacts;
    }

    private static string Concatenate(IEnumerable<string> strings) {
      var sb = new StringBuilder();
      foreach (var s in strings) {
        sb.Append(s);
      }
      return sb.ToString();
    }

    private void ConfigureNodeShapes() {
      graphChart.ClearShapes();
      var font = new Font(FontFamily.GenericSansSerif, 12);
      graphChart.AddShape(typeof(VariableNetworkNode), new LabeledPrimitive(new Ellipse(graphChart.Chart, new PointD(0, 0), new PointD(30, 30), Pens.Black, Brushes.White), "", font));
      graphChart.AddShape(typeof(JunctionNetworkNode), new LabeledPrimitive(new Rectangle(graphChart.Chart, new PointD(0, 0), new PointD(15, 15), Pens.Black, Brushes.DarkGray), "", font));
    }

    #region events
    protected override void OnContentChanged() {
      base.OnContentChanged();
      var run = Content.First();
      var pd = (IRegressionProblemData)run.Parameters["ProblemData"];
      var variables = new HashSet<string>(new List<string>(pd.Dataset.DoubleVariables));
      impactResultNameComboBox.Items.Clear();
      foreach (var result in run.Results.Where(x => x.Value is DoubleMatrix)) {
        var m = (DoubleMatrix)result.Value;
        if (m.RowNames.All(x => variables.Contains(x)))
          impactResultNameComboBox.Items.Add(result.Key);
      }
      qualityResultNameComboBox.Items.Clear();
      foreach (var result in run.Results.Where(x => x.Value is DoubleValue)) {
        qualityResultNameComboBox.Items.Add(result.Key);
      }
      if (impactResultNameComboBox.Items.Count > 0) {
        impactResultNameComboBox.Text = (string)impactResultNameComboBox.Items[0];
      }
      if (qualityResultNameComboBox.Items.Count > 0) {
        qualityResultNameComboBox.Text = (string)qualityResultNameComboBox.Items[0];
      }
      if (impactResultNameComboBox.Items.Count > 0 && qualityResultNameComboBox.Items.Count > 0)
        NetworkConfigurationChanged(this, EventArgs.Empty);
    }

    private void TextBoxValidating(object sender, CancelEventArgs e) {
      double v;
      string errorMsg = "Could not parse the entered value. Please input a real number.";
      var tb = (TextBox)sender;
      if (!double.TryParse(tb.Text, out v)) {
        e.Cancel = true;
        tb.Select(0, tb.Text.Length);

        // Set the ErrorProvider error with the text to display. 
        this.errorProvider.SetError(tb, errorMsg);
        errorProvider.BlinkStyle = ErrorBlinkStyle.NeverBlink;
        errorProvider.SetIconPadding(tb, -20);
      }
    }

    private void NetworkConfigurationBoxValidated(object sender, EventArgs e) {
      var tb = (TextBox)sender;
      errorProvider.SetError(tb, string.Empty);
      NetworkConfigurationChanged(sender, e);
    }

    private void LayoutConfigurationBoxValidated(object sender, EventArgs e) {
      var tb = (TextBox)sender;
      errorProvider.SetError(tb, string.Empty);
      LayoutConfigurationChanged(sender, e);
    }

    private void NetworkConfigurationChanged(object sender, EventArgs e) {
      var useBest = impactAggregationComboBox.SelectedIndex <= 0;
      var threshold = double.Parse(impactThresholdTextBox.Text);
      var qualityResultName = qualityResultNameComboBox.Text;
      var impactsResultName = impactResultNameComboBox.Text;
      if (string.IsNullOrEmpty(qualityResultName) || string.IsNullOrEmpty(impactsResultName))
        return;
      var maximization = maximizationCheckBox.Checked;
      var network = BuildNetworkFromVariableImpacts(Content, qualityResultName, maximization, impactsResultName, threshold, useBest);
      if (network.Vertices.Any())
        graphChart.Graph = network;
    }

    private void LayoutConfigurationChanged(object sender, EventArgs e) {
      ConstrainedForceDirectedLayout.EdgeRouting routingMode;
      switch (edgeRoutingComboBox.SelectedIndex) {
        case 0:
          routingMode = ConstrainedForceDirectedLayout.EdgeRouting.None;
          break;
        case 1:
          routingMode = ConstrainedForceDirectedLayout.EdgeRouting.Polyline;
          break;
        case 2:
          routingMode = ConstrainedForceDirectedLayout.EdgeRouting.Orthogonal;
          break;
        default:
          throw new ArgumentException("Invalid edge routing mode.");
      }
      var idealEdgeLength = double.Parse(idealEdgeLengthTextBox.Text);
      if (routingMode == graphChart.RoutingMode && idealEdgeLength.IsAlmost(graphChart.IdealEdgeLength)) return;
      graphChart.RoutingMode = routingMode;
      graphChart.PerformEdgeRouting = routingMode != ConstrainedForceDirectedLayout.EdgeRouting.None;
      graphChart.IdealEdgeLength = idealEdgeLength;
      graphChart.Draw();
    }
    #endregion
  }
}
